import numpy as np
from .. import combinations as cc_comb
from .. import array as cc_array
# TODO: remove cc_array.advanced_indexing


def reorder_trials(
        trials,
        axis_variables_table=-1,
        axis_samples=-2,
        independent_variables_table=0,
        dtype=None):

    if dtype is None:
        dtype = trials.dtype

    shape_trials = np.asarray(trials.shape, dtype=int)
    n_axes = len(shape_trials)
    axis_variables_table %= n_axes
    axis_samples %= n_axes
    n_trials_per_table = shape_trials[axis_samples]

    n_variables_table = shape_trials[axis_variables_table]
    variables_per_table = np.arange(n_variables_table)

    try:
        len(independent_variables_table)
    except TypeError:
        independent_variables_table = [independent_variables_table]
    independent_variables_table = np.asarray(independent_variables_table, dtype=int)
    n_independent_variables_table = len(independent_variables_table)

    axes = np.arange(n_axes)
    axes_table = np.sort([axis_samples, axis_variables_table])
    variables_axes = axes[np.logical_not(cc_array.samples_in_arr1_are_in_arr2(axes, axes_table))]
    variables_axes_inverted = variables_axes[::-1]

    shape_trials_ordered = shape_trials
    trials_ordered = np.empty(shape_trials_ordered, dtype=dtype)

    indexes_array_independent_variables = np.full(n_axes, 0, dtype=object)
    indexes_array_independent_variables[axis_variables_table] = independent_variables_table
    indexes_array_independent_variables[axis_samples] = np.arange(shape_trials[axis_samples])
    array_independent_variables_1_case = trials[cc_array.advanced_indexing(indexes_array_independent_variables)]

    for a in variables_axes_inverted:
        array_independent_variables_1_case = np.squeeze(array_independent_variables_1_case, axis=a)

    axis_variables_in_combinations = int(axis_variables_table > axis_samples)
    axis_combinations_in_combinations = int(not (bool(axis_variables_in_combinations)))

    conditions_independent_variables = cc_comb.trials_to_conditions(
        array_independent_variables_1_case, axis_combinations=axis_combinations_in_combinations)
    combinations_independent_variables = cc_comb.conditions_to_combinations(
        conditions_independent_variables,
        axis_combinations=axis_combinations_in_combinations)
    n_combinations_independent_variables = combinations_independent_variables.shape[axis_combinations_in_combinations]

    combinations_variables_axes = cc_comb.n_conditions_to_combinations(shape_trials[variables_axes])
    n_combinations_variables_axes = len(combinations_variables_axes)

    indexes_trials = np.empty(n_axes, dtype=object)
    indexes_trials[axis_variables_table] = variables_per_table

    indexes_trials_ordered = np.empty(n_axes, dtype=object)
    indexes_trials_ordered[axis_variables_table] = variables_per_table

    indexes_combinations_independent_variables = np.empty(2, dtype=object)
    indexes_combinations_independent_variables[axis_variables_in_combinations] = np.arange(n_independent_variables_table)

    if n_combinations_variables_axes == 0:
        t = 0
        for c in range(n_combinations_independent_variables):
            indexes_combinations_independent_variables[axis_combinations_in_combinations] = c

            indexes_trials[axis_samples] = np.all(
                array_independent_variables_1_case ==
                combinations_independent_variables[
                    cc_array.advanced_indexing(indexes_combinations_independent_variables)],
                axis=axis_variables_in_combinations)

            k = np.sum(indexes_trials[axis_samples])
            indexes_trials_ordered[axis_samples] = np.arange(t, t + k)
            trials_ordered[cc_array.advanced_indexing(indexes_trials_ordered)] = (
                trials[cc_array.advanced_indexing(indexes_trials)])
            t += k
    else:
        indexes_array_independent_variables = np.empty(n_axes, dtype=object)
        indexes_array_independent_variables[axis_samples] = np.arange(n_trials_per_table)
        indexes_array_independent_variables[axis_variables_table] = independent_variables_table
        for i in range(n_combinations_variables_axes):

            indexes_array_independent_variables[variables_axes] = combinations_variables_axes[i]

            array_independent_variables_i = trials[cc_array.advanced_indexing(indexes_array_independent_variables)]
            for a in variables_axes_inverted:
                array_independent_variables_i = np.squeeze(array_independent_variables_i, axis=a)

            indexes_trials[variables_axes] = combinations_variables_axes[i]
            indexes_trials_ordered[variables_axes] = combinations_variables_axes[i]

            t = 0
            for c in range(n_combinations_independent_variables):

                indexes_combinations_independent_variables[axis_combinations_in_combinations] = c

                indexes_trials[axis_samples] = np.all(
                    array_independent_variables_i ==
                    combinations_independent_variables[
                        cc_array.advanced_indexing(indexes_combinations_independent_variables)],
                    axis=axis_variables_in_combinations)

                k = np.sum(indexes_trials[axis_samples])
                indexes_trials_ordered[axis_samples] = np.arange(t, t + k)
                trials_ordered[cc_array.advanced_indexing(indexes_trials_ordered)] = (
                    trials[cc_array.advanced_indexing(indexes_trials)])
                t += k

    return trials_ordered






import inspect
import numpy as np
import string
import random
from .combinations import n_conditions_to_combinations_on_the_fly as cc_n_conditions_to_combinations_on_the_fly


def n_to_string_of_n_random_letters(n=1):
    # n is the number of random letters

    letters = string.ascii_letters
    string_of_n_random_letters = ''.join(random.choice(letters) for i in range(n))

    return string_of_n_random_letters


def n_to_string_of_n_random_uppercase_letters(n=1):
    # n is the number of random uppercase letters

    uppercase_letters = string.ascii_uppercase
    string_of_n_random_uppercase_letters = ''.join(random.choice(uppercase_letters) for i in range(n))

    return string_of_n_random_uppercase_letters


def n_to_string_of_n_random_lowercase_letters(n=1):
    # n is the number of random lowercase letters

    lowercase_letters = string.ascii_lowercase
    string_of_n_random_lowercase_letters = ''.join(random.choice(lowercase_letters) for i in range(n))

    return string_of_n_random_lowercase_letters


def n_to_string_of_n_random_digits(n=1):
    # n is the number of random digits

    letters_and_numbers = ''.join([string.ascii_letters, string.digits])
    string_of_n_random_digits = ''.join(random.choice(letters_and_numbers) for i in range(n))

    return string_of_n_random_digits


def generate_a_random_email(range_1=None, range_2=None, end='.com'):
    # range_1 is the range of the random n of the email.
    # n is the number of the random digits before "@" of the email.
    # range_2 is the range of the random m of the email.
    # m is the number of the random digits between "@" and ".com" of the email.

    if range_1 is None:
        range_1 = [10, 15]

    if range_2 is None:
        range_2 = [5, 10]

    at = '@'
    len_first_letter = 1
    n_minus_1st_letter = random.randrange(range_1[0] - len_first_letter, range_1[1] - len_first_letter)

    m = random.randrange(range_2[0], range_2[1])

    email = ''.join([
        n_to_string_of_n_random_letters(n=len_first_letter),
        n_to_string_of_n_random_digits(n=n_minus_1st_letter), at,
        n_to_string_of_n_random_digits(n=m), end])

    return email


def generate_array_with_random_emails(shape_array=1, range_1=[10, 15], range_2=[5, 10], end='.com'):
    # shape_array is the shape of the array with random emails.
    # range_1 is the range of the random n of the e_th email.
    # n is the number of the random digits before "@" of the e_th email.
    # range_2 is the range of the random m of the e_th email.
    # m is the number of the random digits between "@" and ".com" of the e_th email.

    # letters_lowercase = string.ascii_lowercase
    # letters_uppercase = string.ascii_uppercase
    # numbers = string.digits
    # letters = ''.join([letters_lowercase, letters_lowercase])
    # letters_and_numbers = ''.join([string.ascii_letters, string.digits])

    at = '@'
    len_at = len(at)
    len_end = len(end)
    len_first_letter = 1

    try:
        n_axes_emails = len(shape_array)
        if not isinstance(shape_array, np.ndarray):
            shape_array = np.asarray(shape_array, dtype=int)
    except ValueError:
        shape_array = np.asarray([shape_array], dtype=int)
        n_axes_emails = 1

    n_minus_1st_letter = np.random.random_integers(
        low=(range_1[0] - len_first_letter), high=(range_1[1] - len_first_letter), size=shape_array)
    m = np.random.random_integers(low=(range_2[0]), high=(range_2[1]), size=shape_array)
    lens_emails_without_1st_letter_at_and_end = n_minus_1st_letter + m
    max_len_emails = lens_emails_without_1st_letter_at_and_end.max() + len_first_letter + len_at + len_end
    dtype = 'U' + str(max_len_emails)

    emails = np.empty(shape_array, dtype=dtype)

    for index_i in cc_n_conditions_to_combinations_on_the_fly(shape_array):
        # print(index_i)
        emails[tuple(index_i)] = ''.join([
            n_to_string_of_n_random_letters(n=len_first_letter),
            n_to_string_of_n_random_digits(n=n_minus_1st_letter[tuple(index_i)]), at,
            n_to_string_of_n_random_digits(n=m[tuple(index_i)]), end])

    return emails


def generate_a_random_name(range_m=[5, 10]):
    # range_m is the range of the random m of the name.
    # m is the number of the random letters of the name.

    len_first_letter = 1
    m_minus_1st_letter = random.randrange(range_m[0] - len_first_letter, range_m[1] - len_first_letter)

    name = ''.join([
        n_to_string_of_n_random_uppercase_letters(n=len_first_letter),
        n_to_string_of_n_random_lowercase_letters(n=m_minus_1st_letter)])

    return name


def generate_array_with_random_names(shape_array=1, range_m=[5, 10]):
    # n is the number of random names
    # range_m is the range of the random m of the r_th name.
    # m is the number of the random letters of the r_th name.

    try:
        n_axes_names = len(shape_array)
        if not isinstance(shape_array, np.ndarray):
            shape_array = np.asarray(shape_array, dtype=int)
    except ValueError:
        shape_array = np.asarray([shape_array], dtype=int)
        n_axes_names = 1

    len_first_letter = 1
    m_minus_1st_letter = np.random.random_integers(
        low=(range_m[0] - len_first_letter), high=(range_m[1] - len_first_letter), size=shape_array)

    max_len_names = m_minus_1st_letter.max() + len_first_letter
    dtype = 'U' + str(max_len_names)

    names = np.empty(shape_array, dtype=dtype)

    for index_i in cc_n_conditions_to_combinations_on_the_fly(shape_array):
        # print(index_i)
        names[tuple(index_i)] = ''.join([
            n_to_string_of_n_random_uppercase_letters(n=len_first_letter),
            n_to_string_of_n_random_lowercase_letters(n=m_minus_1st_letter)])

    return names


def get_variable_name(variable, n_outer_frames=1):
    dict_variables = inspect.getouterframes(inspect.currentframe())[n_outer_frames].frame.f_locals
    keys, values = list(dict_variables.keys()), list(dict_variables.values())
    for i, v in enumerate(values):
        try:
            if v == variable:
                name = keys[i]
        except ValueError:
            if (v == variable).all():
                name = keys[i]
    return name


def get_list_of_n_variable_names(variables, n_outer_frames=1):

    n_variables = len(variables)
    list_of_names = [None] * n_variables
    max_string_length = 0
    for i, v in enumerate(variables):
        list_of_names[i] = get_variable_name(v, n_outer_frames=n_outer_frames + 1)
        string_length_i = len(list_of_names[i])
        if string_length_i > max_string_length:
            max_string_length = string_length_i
    return np.asarray(list_of_names, dtype=('<U' + str(max_string_length)))


def format_float_to_str(value_float, n_decimals=None):
    if n_decimals is None:
        template = '{:f}'
    else:
        template = '{:.' + str(n_decimals) + 'f}'

    value_str = template.format(value_float)
    return value_str




import numpy as np
from scipy.stats import t
# from scipy.stats.stats import _ttest_finish
from ..combinations import n_conditions_to_combinations, trials_to_n_conditions
from ..array import samples_in_arr1_are_in_arr2


def scores_to_diff_of_scores(scores, axis, delta=1, stride=1, keepdims=False):
    # this function will go to cc.array in future versions

    shape_scores = np.asarray(scores.shape)
    n_axes_scores = shape_scores.size
    if axis < 0:
        axis += n_axes_scores

    n_conditions = shape_scores[axis]
    index_0 = np.empty(n_axes_scores, dtype=object)
    index_0[:] = slice(None)
    index_1 = np.copy(index_0)

    index_diff = np.copy(index_0)
    index_diff[axis] = 0

    n_differences = int((n_conditions - abs(delta)) // abs(stride))
    shape_diff_of_scores = shape_scores
    shape_diff_of_scores[axis] = n_differences
    diff_of_scores = np.empty(shape_diff_of_scores, dtype=scores.dtype)

    if delta > 0 and stride > 0:
        for i in range(0, n_conditions - delta, stride):
            index_0[axis] = i
            index_1[axis] = i + delta
            diff_of_scores[tuple(index_diff)] = scores[tuple(index_0)] - scores[tuple(index_1)]
            index_diff[axis] += 1
    elif delta < 0 and stride > 0:
        for i in range(abs(delta), n_conditions, stride):
            index_0[axis] = i
            index_1[axis] = i + delta
            diff_of_scores[tuple(index_diff)] = scores[tuple(index_0)] - scores[tuple(index_1)]
            index_diff[axis] += 1
    elif delta > 0 and stride < 0:
        for i in range(n_conditions - delta - 1, -1, stride):
            index_0[axis] = i
            index_1[axis] = i + delta
            diff_of_scores[tuple(index_diff)] = scores[tuple(index_0)] - scores[tuple(index_1)]
            index_diff[axis] += 1
    elif delta < 0 and stride < 0:
        for i in range(n_conditions - 1, abs(delta) -1, stride):
            index_0[axis] = i
            index_1[axis] = i + delta
            diff_of_scores[tuple(index_diff)] = scores[tuple(index_0)] - scores[tuple(index_1)]
            index_diff[axis] += 1
    elif delta == 0:
        raise ValueError('delta has to be an intiger smaller or greater than 0. It cannot be 0')
    elif stride == 0:
        raise ValueError('stride has to be an intiger smaller or greater than 0. It cannot be 0')
    else:
        raise ValueError('Both delta and stride have to be an intiger smaller or greater than 0.')

    if not keepdims and (diff_of_scores.shape[axis] == 1):
        diff_of_scores = np.squeeze(diff_of_scores, axis=axis)

    return diff_of_scores


def scores_to_n_samples(scores, axis_samples, keepdims=False):

    if scores.dtype == object:

        shape_object_scores = np.asarray(scores.shape)
        n_axes_object_scores = shape_object_scores.size
        axes_object_scores = np.arange(n_axes_object_scores)

        indexes_object = np.empty(n_axes_object_scores, dtype=object)
        indexes_object[:] = 0
        shape_scores = np.asarray(scores[tuple(indexes_object)].shape)
        n_axes_scores = shape_scores.size
        if axis_samples < 0:
            axis_samples += n_axes_scores

        if keepdims:
            shape_scores_tmp = shape_scores
            shape_scores_tmp[axis_samples] = 1
        else:
            shape_scores_tmp = shape_scores[np.arange(n_axes_scores) != axis_samples]

        shape_n_samples = np.append(shape_object_scores, shape_scores_tmp)
        n_samples = np.empty(shape_n_samples, dtype=int)

        n_axes_n_samples = shape_n_samples.size
        indexes_n_samples = np.empty(n_axes_n_samples, dtype=object)
        indexes_n_samples[:] = slice(None)

        indexes_object = n_conditions_to_combinations(shape_object_scores)
        for indexes_object_i in indexes_object:

            indexes_n_samples[axes_object_scores] = indexes_object_i
            indexes_n_samples_tuple = tuple(indexes_n_samples)
            indexes_object_i_tuple = tuple(indexes_object_i)

            n_samples[indexes_n_samples_tuple] = np.sum(np.logical_not(np.isnan(
                scores[indexes_object_i_tuple])), axis=axis_samples, keepdims=keepdims)

    else:
        shape_scores = np.asarray(scores.shape)
        n_axes_scores = shape_scores.size
        if axis_samples < 0:
            axis_samples += n_axes_scores

        n_samples = np.sum(np.logical_not(np.isnan(scores)), axis=axis_samples, keepdims=keepdims)

    return n_samples


def samples_to_frequencies(data, axis_samples, axis_variables_table):

    if axis_samples < axis_variables_table:
        axis_comb = 0
    elif axis_samples > axis_variables_table:
        axis_comb = 1
    else:
        raise ValueError('The following assumption is not met:\n'
                         '\taxis_samples \u003D axis_variables_table')

    shape_data = np.asarray(data.shape, dtype=int)
    n_axes_data = shape_data.size
    axes_data = np.arange(n_axes_data)
    axes_others_and_samples = axes_data[axes_data != axis_variables_table]
    axes_other_data = axes_others_and_samples[axes_others_and_samples != axis_samples]
    n_axes_other_data = axes_other_data.size
    n_variables_table = shape_data[axis_variables_table]

    indexes_data = np.empty(n_axes_data, dtype=object)
    indexes_data[:] = slice(None)
    indexes_data[axes_other_data] = 0

    n_conditions = trials_to_n_conditions(data[tuple(indexes_data)], axis_combinations=axis_comb)

    n_axes_frequencies = n_axes_other_data + n_variables_table
    shape_frequencies = np.empty(n_axes_frequencies, dtype=int)
    shape_frequencies[slice(n_axes_other_data)] = shape_data[axes_other_data]
    shape_frequencies[slice(n_axes_other_data, n_axes_frequencies)] = n_conditions
    frequencies = np.empty(shape_frequencies, dtype=int)
    frequencies[:] = 0
    indexes_frequencies = np.empty(n_axes_data, dtype=object)

    combinations_others_and_samples = n_conditions_to_combinations(shape_data[axes_others_and_samples])
    # combinations_others = combinations_others_and_samples[slice(None), axes_others_and_samples != axis_samples]
    indexes_comb_c = axes_others_and_samples != axis_samples
    for comb_c in combinations_others_and_samples:

        indexes_data[axes_others_and_samples] = comb_c

        indexes_frequencies[slice(n_axes_other_data)] = comb_c[indexes_comb_c]
        indexes_frequencies[slice(n_axes_other_data, n_axes_frequencies)] = data[tuple(indexes_data)]

        frequencies[tuple(indexes_frequencies)] += 1

    return frequencies


def samples_to_probabilities(data, axis_samples, axis_variables_table):

    if axis_samples == axis_variables_table:
        raise ValueError('The following assumption is not met:\n'
                         '\taxis_samples \u2260 axis_variables_table')
    n_samples = data.shape[axis_samples]
    frequencies = samples_to_frequencies(data, axis_samples, axis_variables_table)
    probabilities = frequencies / n_samples

    return probabilities


def samples_to_percentages(data, axis_samples, axis_variables_table):
    if axis_samples == axis_variables_table:
        raise ValueError('The following assumption is not met:\n'
                         '\taxis_samples \u2260 axis_variables_table')
    n_samples = data.shape[axis_samples]
    frequencies = samples_to_frequencies(data, axis_samples, axis_variables_table)
    percentages = frequencies / n_samples * 100

    return percentages


def scores_to_df_of_variance_and_paired_t(scores, axis_samples, keepdims=False):

    n_samples = scores_to_n_samples(scores, axis_samples=axis_samples, keepdims=keepdims)
    df = n_samples - 1

    return df


def scores_to_means(scores, axes, keepdims=False):

    axes_means = axes

    if scores.dtype == object:

        shape_object_scores = np.asarray(scores.shape)
        n_axes_object_scores = shape_object_scores.size
        axes_object_scores = np.arange(n_axes_object_scores)

        indexes_object = np.empty(n_axes_object_scores, dtype=object)
        indexes_object[:] = 0
        shape_scores = np.asarray(scores[tuple(indexes_object)].shape, dtype=int)
        n_axes_scores = shape_scores.size
        try:
            len(axes_means)
            # n_axes_means = len(axes_means)
            axes_means = np.asarray(axes_means, dtype=int)
            axes_means[axes_means < 0] += n_axes_scores
            # check point 1
            if np.sum(axes_means[0] == axes_means) > 1:
                raise ValueError('axes cannot contain repeated values')
            axes_means = np.sort(axes_means)[::-1]
        except TypeError:
            if axes_means < 0:
                axes_means += n_axes_scores
            axes_means = np.asarray([axes_means], dtype=int)
            # n_axes_means = 1

        if keepdims:
            shape_scores_tmp = shape_scores
            shape_scores_tmp[axes_means] = 1
        else:
            axes_scores = np.arange(n_axes_scores)
            shape_scores_tmp = shape_scores[np.logical_not(samples_in_arr1_are_in_arr2(
                axes_scores, axes_means))]

        shape_means = np.append(shape_object_scores, shape_scores_tmp)
        means = np.empty(shape_means, dtype=float)

        n_axes_means = shape_means.size
        indexes_means = np.empty(n_axes_means, dtype=object)
        indexes_means[:] = slice(None)

        indexes_object = n_conditions_to_combinations(shape_object_scores)
        for indexes_object_i in indexes_object:

            indexes_means[axes_object_scores] = indexes_object_i
            indexes_means_tuple_i = tuple(indexes_means)
            indexes_object_tuple_i = tuple(indexes_object_i)

            scores_tmp_i = scores[indexes_object_tuple_i]
            for a in axes_means:
                n_samples = np.sum(np.logical_not(np.isnan(scores_tmp_i)), axis=a, keepdims=keepdims)
                sum_of_scores = np.nansum(scores_tmp_i, axis=a, keepdims=keepdims)
                scores_tmp_i = sum_of_scores / n_samples

            means[indexes_means_tuple_i] = scores_tmp_i

    else:

        shape_scores = np.asarray(scores.shape, dtype=int)
        n_axes_scores = shape_scores.size

        try:
            len(axes_means)
            # n_axes_means = len(axes_means)
            axes_means = np.asarray(axes_means, dtype=int)
            axes_means[axes_means < 0] += n_axes_scores
            # check point 1
            if np.sum(axes_means[0] == axes_means) > 1:
                raise ValueError('axes cannot contain repeated values')
            axes_means = np.sort(axes_means)[::-1]
        except TypeError:
            if axes_means < 0:
                axes_means += n_axes_scores
            axes_means = np.asarray([axes_means], dtype=int)
            # n_axes_means = 1
        scores_tmp = scores
        for a in axes_means:
            sum_of_scores = np.nansum(scores_tmp, axis=a, keepdims=keepdims)
            n_samples = np.sum(np.logical_not(np.isnan(scores_tmp)), axis=a, keepdims=keepdims)
            scores_tmp = sum_of_scores / n_samples
        means = scores_tmp

    return means


def scores_to_variances(scores, axis_samples, keepdims=False):

    if scores.dtype == object:

        shape_object_scores = np.asarray(scores.shape)
        n_axes_object_scores = shape_object_scores.size
        axes_object_scores = np.arange(n_axes_object_scores)

        indexes_object = np.empty(n_axes_object_scores, dtype=object)
        indexes_object[:] = 0
        shape_scores = np.asarray(scores[tuple(indexes_object)].shape)
        n_axes_scores = shape_scores.size
        if axis_samples < 0:
            axis_samples += n_axes_scores

        if keepdims:
            shape_scores_tmp = shape_scores
            shape_scores_tmp[axis_samples] = 1
        else:
            shape_scores_tmp = shape_scores[np.arange(n_axes_scores) != axis_samples]

        shape_variances = np.append(shape_object_scores, shape_scores_tmp)
        variances = np.empty(shape_variances, dtype=float)

        n_axes_variances = shape_variances.size
        indexes_variances = np.empty(n_axes_variances, dtype=object)
        indexes_variances[:] = slice(None)

        indexes_object = n_conditions_to_combinations(shape_object_scores)
        for indexes_object_i in indexes_object:

            indexes_variances[axes_object_scores] = indexes_object_i
            indexes_variances_tuple = tuple(indexes_variances)
            indexes_object_i_tuple = tuple(indexes_object_i)

            n_samples = np.sum(np.logical_not(np.isnan(
                scores[indexes_object_i_tuple])), axis=axis_samples, keepdims=True)
            sum_of_scores = np.nansum(scores[indexes_object_i_tuple], axis=axis_samples, keepdims=True)
            means = sum_of_scores / n_samples

            sum_of_squared_distances = np.nansum(
                (scores[indexes_object_i_tuple] - means) ** 2,
                axis=axis_samples, keepdims=keepdims)

            if not keepdims:
                n_samples = np.squeeze(n_samples, axis=axis_samples)
            df = n_samples - 1

            variances[indexes_variances_tuple] = sum_of_squared_distances / df

    else:

        shape_scores = np.asarray(scores.shape)
        n_axes_scores = shape_scores.size
        if axis_samples < 0:
            axis_samples += n_axes_scores

        n_samples = np.sum(np.logical_not(np.isnan(scores)), axis=axis_samples, keepdims=True)
        means = np.nansum(scores, axis=axis_samples, keepdims=True) / n_samples
        sum_of_squared_distances = np.nansum((scores - means) ** 2, axis=axis_samples, keepdims=keepdims)
        if not keepdims:
            n_samples = np.squeeze(n_samples, axis=axis_samples)
        df = n_samples - 1
        variances = sum_of_squared_distances / df

    return variances


def scores_to_standard_deviations(scores, axis_samples, keepdims=False):

    variances = scores_to_variances(scores, axis_samples, keepdims=keepdims)
    standard_deviations = np.sqrt(variances)

    return standard_deviations


def scores_to_standard_errors(scores, axis_samples, keepdims=False):

    # shape_scores = np.asarray(scores.shape)
    # n_axes_scores = shape_scores.size
    # if axis_samples < 0:
    #     axis_samples += n_axes_scores

    n_samples = scores_to_n_samples(scores, axis_samples, keepdims=keepdims)
    variances = scores_to_variances(scores, axis_samples, keepdims=keepdims)
    std_error = np.sqrt(variances / n_samples)

    return std_error


def scores_to_confidence_intervals(
        scores, axis_samples, alpha=0.05, tails='2', keepdims=False):

    confidence = 1 - alpha

    # shape_scores = np.asarray(scores.shape)
    # n_axes_scores = shape_scores.size
    # if axis_samples < 0:
    #     axis_samples += n_axes_scores

    std_err = scores_to_standard_errors(scores, axis_samples, keepdims=keepdims)
    df = scores_to_df_of_variance_and_paired_t(scores, axis_samples, keepdims=keepdims)

    shape_df = np.asarray(df.shape)
    indexes_df = n_conditions_to_combinations(shape_df)
    t_critical = np.empty(shape_df, dtype=float)

    for indexes_df_i in indexes_df:

        indexes_df_i_tuple = tuple(indexes_df_i)

        if tails == '2':
            t_critical[indexes_df_i_tuple] = t.ppf((1 + confidence) / 2., df[indexes_df_i_tuple])

        elif tails == '1l':
            t_critical[indexes_df_i_tuple] = -t.ppf(confidence, df[indexes_df_i_tuple])
        elif tails == '1r':
            t_critical[indexes_df_i_tuple] = t.ppf(confidence, df[indexes_df_i_tuple])

    h = std_err * t_critical

    return h



import numpy as np
import math


def rint(num):
    if isinstance(num, int):
        return num
    elif str(num).__contains__('.'):
        int_down = math.floor(num)
        int_up = math.ceil(num)
        if int_down == int_up:
            return int_down
        elif num >= (int_down + 0.5):
            return int_up
        else:
            return int_down
    else:
        return int(num)


def convert_to_int_or_float(num):
    if isinstance(num, int):
        return num
    else:
        str_num = str(num)
        list_num = str_num.split('.')
        len_num = len(list_num)
        if len_num == 1:
            return int(num)
        elif len_num == 2:
            list_decimal = list_num[1].split('e')
            len_decimal = len(list_decimal)
            if len_decimal == 1:
                if int(list_num[1]) == 0:
                    return int(num)
                elif isinstance(num, float):
                    return num
                else:
                    return float(num)
            elif int(list_decimal[1]) < 0:
                if isinstance(num, float):
                    return num
                else:
                    return float(num)
            elif int(list_decimal[1]) > 0:
                return rint(num)

        else:
            raise TypeError('num')


def if_equal_to_the_nearest_int_convert_to_int_else_float(num):
    if isinstance(num, int):
        return num
    else:
        num_rounded = rint(num)
        if num_rounded == num:
            return num_rounded
        else:
            return float(num)


def factors_of_x(x, y=1):

    x = convert_to_int_or_float(x)
    y = convert_to_int_or_float(y)

    if isinstance(x, float):
        raise TypeError('Type of x must be int')

    if isinstance(y, float):
        raise TypeError('Type of y must be int')

    factors = np.empty(0, dtype='i')

    for i in range(y, x + 1):
        if x % i == 0:
            factors = np.append(factors, i)

    return factors


def prod(numbers, start=1):

    product = convert_to_int_or_float(start)
    for num_i in numbers:
        product *= convert_to_int_or_float(num_i)
    return product


def gamma(z):
    
    print('scipy.special.gamma(z) is more efficient')
    
    pos_inf = 100
    n_dx = 10000000
    dx = pos_inf / n_dx

    if not isinstance(z, np.ndarray):
        z = np.asarray(z)

    n_axes_z = len(z.shape)
    n_axes_x = n_axes_z + 1
    axis_delta = n_axes_x - 1
    x = np.arange(0, pos_inf, dx)
    while len(x.shape) < n_axes_x:
        x = np.expand_dims(x, axis=0)

    if n_axes_z > 0:
        z = np.expand_dims(z, axis_delta)

    with np.errstate(divide='ignore'):
        return np.sum(
            np.power(x, z - 1) * np.power(math.e, -x) * dx, axis=axis_delta)



import numpy as np
from ...array import samples_in_arr1_are_in_arr2  # , advanced_indexing
from ...combinations import (
    trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations)

# def from_2_arrays_new(
#         array_variables_staying,
#         array_variables_adding_axes,
#         axis_samples,
#         axis_variables_table,
#         axes_inserting=None,
#         dtype=None,
#         format_and_check=True):
#
#     axis_variables_table_input = axis_variables_table
#     axis_samples_input = axis_samples
#
#     shape_array_variables_staying = np.asarray(array_variables_staying.shape)
#     n_axes_array_input = shape_array_variables_staying.size
#
#     if format_and_check:
#         n_axes_array_variables_staying = n_axes_array_input
#
#         shape_array_variables_adding_axes = np.asarray(array_variables_adding_axes.shape)
#         n_axes_array_variables_adding_axes = shape_array_variables_adding_axes.size
#
#         # check point 1
#         if n_axes_array_variables_staying != n_axes_array_variables_adding_axes:
#             raise ValueError('dimension mismatch')
#
#         if axis_variables_table_input < 0:
#             axis_variables_table_input += n_axes_array_input
#         if axis_samples_input < 0:
#             axis_samples_input += n_axes_array_input
#
#         # check point 5
#         if axis_samples_input == axis_variables_table_input:
#             raise ValueError('axis_samples_input and axis_variables_table_input must be different')
#
#         # check point 2
#         indexes_logical = np.arange(n_axes_array_input) != axis_variables_table_input
#         if np.any(shape_array_variables_staying[indexes_logical] != shape_array_variables_adding_axes[indexes_logical]):
#             raise ValueError('dimension mismatch')
#
#         n_variables_table_adding_axes = shape_array_variables_adding_axes[axis_variables_table_input]
#
#         # format axes_inserting
#         n_axes_inserting = n_variables_table_adding_axes
#         n_axes_array_output_object = n_axes_inserting
#         try:
#             n_axes_inserting = len(axes_inserting)
#             axes_inserting = np.asarray(axes_inserting, dtype=int)
#             axes_inserting[axes_inserting < 0] += n_axes_array_output_object
#             # check point 3
#             if np.sum(axes_inserting[0] == axes_inserting) > 1:
#                 raise ValueError('axes_inserting cannot contain repeated values')
#             # check point 4
#             if n_variables_table_adding_axes != n_axes_inserting:
#                 raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')
#         except TypeError:
#             if axes_inserting is None:
#                 axes_inserting = np.arange(n_axes_inserting)
#             else:
#                 if axes_inserting < 0:
#                     axes_inserting += n_axes_array_output_object
#                 axes_inserting = np.asarray([axes_inserting], dtype=int)
#                 n_axes_inserting = 1
#                 # check point 4
#                 if n_variables_table_adding_axes != n_axes_inserting:
#                     raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')
#
#         if dtype is not None:
#             array_variables_staying = array_variables_staying.astype(dtype)
#
#     else:
#         n_axes_inserting = axes_inserting.size
#         n_axes_array_output_object = n_axes_inserting
#
#     # n_variables_table_staying = shape_array_variables_staying[axis_variables_table_input]
#
#     axes_array_input = np.arange(n_axes_array_input)
#     axes_non_axis_variables_table_input = axes_array_input[axes_array_input != axis_variables_table_input]
#     axes_other_array_input = axes_non_axis_variables_table_input[
#         axes_non_axis_variables_table_input != axis_samples_input]
#     n_axes_other_array_input = axes_other_array_input.size
#
#     indexes_array_variables_staying = np.empty(n_axes_array_input, dtype=object)
#     indexes_array_variables_staying[:] = slice(None)
#
#     axis_variables_in_combinations = int(axis_variables_table_input > axis_samples_input)
#     axis_combinations_in_combinations = int(not (bool(axis_variables_in_combinations)))
#
#     if n_axes_other_array_input == 0:
#         conditions_variables_table_adding_axes = trials_to_conditions(
#             array_variables_adding_axes, axis_combinations=axis_combinations_in_combinations)
#     elif n_axes_other_array_input > 0:
#         indexes_array_variables_adding_axes = np.copy(indexes_array_variables_staying)
#         indexes_array_variables_adding_axes[axes_other_array_input] = 0
#         array_variables_adding_axes_1 = array_variables_adding_axes[tuple(indexes_array_variables_adding_axes)]
#         conditions_variables_table_adding_axes = trials_to_conditions(
#             array_variables_adding_axes_1, axis_combinations=axis_combinations_in_combinations)
#
#     n_conditions_variables_table_adding_axes = conditions_to_n_conditions(
#         conditions_variables_table_adding_axes)
#     n_combinations_variables_table_adding_axes = int(np.prod(
#         n_conditions_variables_table_adding_axes))
#
#     combinations_variables_table_adding_axes = conditions_to_combinations(
#         conditions_variables_table_adding_axes,
#         axis_combinations=axis_combinations_in_combinations)
#
#     combinations_axes_inserting = n_conditions_to_combinations(
#         n_conditions_variables_table_adding_axes)
#
#     # axes_array_output_object = np.arange(n_axes_array_output_object)
#     shape_array_output_object = np.empty(n_axes_array_output_object, dtype=int)
#     shape_array_output_object[axes_inserting] = n_conditions_variables_table_adding_axes
#     array_output_object = np.empty(shape_array_output_object, dtype=object)
#
#     indexes_output_object = np.empty(n_axes_array_output_object, dtype=object)
#
#     indexes_combinations = np.empty(2, dtype=object)
#     indexes_combinations[axis_variables_in_combinations] = slice(None)
#
#     if n_axes_other_array_input == 0:
#         for i in range(n_combinations_variables_table_adding_axes):
#             indexes_output_object[axes_inserting] = combinations_axes_inserting[i]
#             indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
#             indexes_array_variables_staying[axis_samples_input] = np.all(
#                 array_variables_adding_axes ==
#                 combinations_variables_table_adding_axes[tuple(indexes_combinations)],
#                 axis=axis_variables_in_combinations)
#             array_output_object[tuple(indexes_output_object)] = array_variables_staying[tuple(
#                 indexes_array_variables_staying)]
#
#     elif n_axes_other_array_input > 0:
#
#         shape_array_output = np.copy(shape_array_variables_staying)
#         shape_array_output[axis_samples_input] = (
#                 shape_array_variables_staying[axis_samples_input] / n_combinations_variables_table_adding_axes)
#
#         dtype = array_variables_staying.dtype
#         array_output = np.empty(shape_array_output, dtype=dtype)
#
#         n_array_output_object_tmp = 1
#         array_output_object_tmp = np.empty(n_array_output_object_tmp, dtype=object)
#         array_output_object_tmp[0] = array_output
#         while n_array_output_object_tmp < n_axes_array_output_object:
#             array_output_object_tmp = np.expand_dims(array_output_object_tmp, axis=0)
#             n_array_output_object_tmp = len(array_output_object_tmp.shape)
#         array_output_object[:] = array_output_object_tmp
#
#         indexes_output = np.empty(n_axes_array_input, dtype=object)
#         indexes_output[:] = slice(None)
#
#         indexes_array_variables_staying_j = np.empty(2, dtype=object)
#         indexes_array_variables_staying_j[axis_variables_in_combinations] = slice(None)
#
#         combinations_axes_other = n_conditions_to_combinations(shape_array_variables_staying[axes_other_array_input])
#         n_combinations_axes_other = combinations_axes_other.shape[0]
#         for i in range(n_combinations_variables_table_adding_axes):
#
#             indexes_output_object[axes_inserting] = combinations_axes_inserting[i]
#             array_output_object[tuple(indexes_output_object)] = np.copy(array_output_object[tuple(indexes_output_object)])
#
#             indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
#
#             for j in range(n_combinations_axes_other):
#                 combinations_axes_other_j = combinations_axes_other[j]
#                 indexes_output[axes_other_array_input] = combinations_axes_other_j
#
#                 indexes_array_variables_staying[axes_other_array_input] = combinations_axes_other_j
#                 indexes_array_variables_adding_axes[axes_other_array_input] = combinations_axes_other_j
#
#                 array_variables_adding_axes_j = array_variables_adding_axes[
#                     tuple(indexes_array_variables_adding_axes)]
#                 array_variables_staying_j = array_variables_staying[tuple(
#                     indexes_array_variables_staying)]
#
#                 indexes_array_variables_staying_j[axis_combinations_in_combinations] = np.all(
#                     array_variables_adding_axes_j ==
#                     combinations_variables_table_adding_axes[tuple(indexes_combinations)],
#                     axis=axis_variables_in_combinations)
#                 array_output_object[tuple(indexes_output_object)][tuple(indexes_output)] = array_variables_staying_j[
#                     tuple(indexes_array_variables_staying_j)]
#
#             # indexes_output_object[axes_inserting] = combinations_axes_inserting[i]
#             # array_output_object[tuple(indexes_output_object)] = np.copy(array_output)
#
#     return array_output_object


def from_2_arrays(
        array_variables_staying,
        array_variables_adding_axes,
        axis_samples,
        axis_variables_table,
        axes_inserting=None,
        dtype=None,
        format_and_check=True):

    axis_variables_table_input = axis_variables_table
    axis_samples_input = axis_samples

    shape_array_variables_staying = np.asarray(array_variables_staying.shape)
    n_axes_array_input = shape_array_variables_staying.size
    # n_variables_table_staying = shape_array_variables_staying[axis_variables_table_input]

    if format_and_check:
        n_axes_array_variables_staying = n_axes_array_input

        shape_array_variables_adding_axes = np.asarray(array_variables_adding_axes.shape)
        n_axes_array_variables_adding_axes = shape_array_variables_adding_axes.size

        # check point 1
        if n_axes_array_variables_staying != n_axes_array_variables_adding_axes:
            raise ValueError('dimension mismatch')

        if axis_variables_table_input < 0:
            axis_variables_table_input += n_axes_array_input
        if axis_samples_input < 0:
            axis_samples_input += n_axes_array_input

        # check point 5
        if axis_samples_input == axis_variables_table_input:
            raise ValueError('axis_samples_input and axis_variables_table_input must be different')

        # check point 2
        indexes_logical = np.arange(n_axes_array_input) != axis_variables_table_input
        if np.any(shape_array_variables_staying[indexes_logical] != shape_array_variables_adding_axes[indexes_logical]):
            raise ValueError('dimension mismatch')

        n_variables_table_adding_axes = shape_array_variables_adding_axes[axis_variables_table_input]

        # format axes_inserting
        n_axes_inserting = n_variables_table_adding_axes
        n_axes_array_output_object = n_axes_inserting
        try:
            len(axes_inserting)
            axes_inserting = np.asarray(axes_inserting, dtype=int)
            axes_inserting[axes_inserting < 0] += n_axes_array_output_object
            # check point 3
            if np.sum(axes_inserting[0] == axes_inserting) > 1:
                raise ValueError('axes_inserting cannot contain repeated values')

        except TypeError:
            if axes_inserting is None:
                axes_inserting = np.arange(n_axes_inserting)
            else:
                if axes_inserting < 0:
                    axes_inserting += n_axes_array_output_object
                axes_inserting = np.asarray([axes_inserting], dtype=int)

        if dtype is not None:
            array_variables_staying = array_variables_staying.astype(dtype)

    else:
        n_axes_inserting = axes_inserting.size
        n_axes_array_output_object = n_axes_inserting

    axes_array_input = np.arange(n_axes_array_input)
    axes_non_axis_variables_table_input = axes_array_input[axes_array_input != axis_variables_table_input]
    axes_other_array_input = axes_non_axis_variables_table_input[
        axes_non_axis_variables_table_input != axis_samples_input]
    n_axes_other_array_input = axes_other_array_input.size

    indexes_array_variables_staying = np.empty(n_axes_array_input, dtype=object)
    indexes_array_variables_staying[:] = slice(None)

    axis_variables_in_combinations = int(axis_variables_table_input > axis_samples_input)
    axis_combinations_in_combinations = int(not (bool(axis_variables_in_combinations)))

    if n_axes_other_array_input == 0:
        conditions_variables_table_adding_axes = trials_to_conditions(
            array_variables_adding_axes, axis_combinations=axis_combinations_in_combinations)
    elif n_axes_other_array_input > 0:
        indexes_array_variables_adding_axes = np.copy(indexes_array_variables_staying)
        indexes_array_variables_adding_axes[axes_other_array_input] = 0
        array_variables_adding_axes_1 = array_variables_adding_axes[tuple(indexes_array_variables_adding_axes)]
        conditions_variables_table_adding_axes = trials_to_conditions(
            array_variables_adding_axes_1, axis_combinations=axis_combinations_in_combinations)

    n_conditions_variables_table_adding_axes = conditions_to_n_conditions(
        conditions_variables_table_adding_axes)
    n_combinations_variables_table_adding_axes = int(np.prod(
        n_conditions_variables_table_adding_axes))

    combinations_variables_table_adding_axes = conditions_to_combinations(
        conditions_variables_table_adding_axes,
        axis_combinations=axis_combinations_in_combinations)

    combinations_axes_inserting = n_conditions_to_combinations(
        n_conditions_variables_table_adding_axes)

    # axes_array_output_object = np.arange(n_axes_array_output_object)
    shape_array_output_object = np.empty(n_axes_array_output_object, dtype=int)
    shape_array_output_object[axes_inserting] = n_conditions_variables_table_adding_axes
    array_output_object = np.empty(shape_array_output_object, dtype=object)

    indexes_output_object = np.empty(n_axes_array_output_object, dtype=object)

    indexes_combinations = np.empty(2, dtype=object)
    indexes_combinations[axis_variables_in_combinations] = slice(None)

    if n_axes_other_array_input == 0:
        for i in range(n_combinations_variables_table_adding_axes):
            indexes_output_object[axes_inserting] = combinations_axes_inserting[i]
            indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
            indexes_array_variables_staying[axis_samples_input] = np.all(
                array_variables_adding_axes ==
                combinations_variables_table_adding_axes[tuple(indexes_combinations)],
                axis=axis_variables_in_combinations)
            array_output_object[tuple(indexes_output_object)] = array_variables_staying[tuple(
                indexes_array_variables_staying)]

    elif n_axes_other_array_input > 0:

        shape_array_output = np.copy(shape_array_variables_staying)
        shape_array_output[axis_samples_input] = (
                shape_array_variables_staying[axis_samples_input] / n_combinations_variables_table_adding_axes)

        dtype = array_variables_staying.dtype
        array_output = np.empty(shape_array_output, dtype=dtype)

        for i in range(n_combinations_variables_table_adding_axes):
            indexes_output_object[axes_inserting] = combinations_axes_inserting[i]
            array_output_object[tuple(indexes_output_object)] = np.copy(array_output)

        indexes_output = np.empty(n_axes_array_input, dtype=object)
        indexes_output[:] = slice(None)

        indexes_array_variables_staying_j = np.empty(2, dtype=object)
        indexes_array_variables_staying_j[axis_variables_in_combinations] = slice(None)

        combinations_axes_other = n_conditions_to_combinations(shape_array_variables_staying[axes_other_array_input])
        n_combinations_axes_other = combinations_axes_other.shape[0]
        for j in range(n_combinations_axes_other):
            combinations_axes_other_j = combinations_axes_other[j]
            indexes_output[axes_other_array_input] = combinations_axes_other_j
            indexes_output_tuple = tuple(indexes_output)
            indexes_array_variables_staying[axes_other_array_input] = combinations_axes_other_j
            indexes_array_variables_adding_axes[axes_other_array_input] = combinations_axes_other_j
            array_variables_adding_axes_j = array_variables_adding_axes[
                tuple(indexes_array_variables_adding_axes)]
            array_variables_staying_j = array_variables_staying[tuple(
                indexes_array_variables_staying)]
            for i in range(n_combinations_variables_table_adding_axes):
                indexes_output_object[axes_inserting] = combinations_axes_inserting[i]
                indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
                indexes_array_variables_staying_j[axis_combinations_in_combinations] = np.all(
                    array_variables_adding_axes_j ==
                    combinations_variables_table_adding_axes[tuple(indexes_combinations)],
                    axis=axis_variables_in_combinations)
                array_output_object[tuple(indexes_output_object)][indexes_output_tuple] = array_variables_staying_j[
                    tuple(indexes_array_variables_staying_j)]

    return array_output_object


def from_2_arrays_advanced(
        array_variables_staying,
        array_variables_adding_axes,
        axis_samples,
        axis_variables_table,
        variables_table_adding_axes=None,
        axes_inserting=None,
        variables_table_staying=None,
        dtype=None,
        format_and_check=True):

    # Notes:
    # 1) uses the efficient function slice for indexes;
    # 2) it splits the array in two arrays first: array_variables_adding_axes and array_variables_staying.
    #    Then, it makes the array output from those two arrays;
    # 3) it does not assumes that the order of samples (or trials) in the axis_samples are is the same
    #    in each combination of variables' axes. The order is based on the order of the combinations
    #    of variables' conditions in axis_variables_table. In other words, the samples can be in random order
    #    (or in different order) in each combinations of variables axes/conditions.

    # Input requirements:
    # 1) the numbers of axes of the two arrays must be equal;
    # 2) the shapes of the 2 arrays must be equal, except for the axis_variable_table,
    #    i.e. the axis that contains the variables of the tables;
    # 3) axes_inserting cannot contain repeated values;
    # 4) variables_table_adding_axes cannot contain repeated values;
    # 5) shapes of axes_inserting and variables_table_adding_axes must be equal
    # 6) axis_samples != axis_variables_table
    # 7) the numbers of samples (or trials) in the axis_samples for all combinations
    #    of variables_table_adding_axes are equal

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.combinations import trials_to_conditions
    # 4) from ccalafiore.combinations import conditions_to_n_conditions
    # 5) from ccalafiore.combinations import conditions_to_combinations
    # 6) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # from ccalafiore.combinations import \
    #     trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations
    # from ccalafiore.array import samples_in_arr1_are_in_arr2

    axis_variables_table_input = axis_variables_table
    axis_samples_input = axis_samples

    shape_array_variables_staying = np.asarray(array_variables_staying.shape)
    n_axes_array_variables_staying = shape_array_variables_staying.size
    n_variables_table_in_array_variables_staying = shape_array_variables_staying[axis_variables_table_input]
    variables_table_in_array_variables_staying = np.arange(n_variables_table_in_array_variables_staying)

    shape_array_variables_adding_axes = np.asarray(array_variables_adding_axes.shape)
    n_axes_array_variables_adding_axes = shape_array_variables_adding_axes.size
    n_variables_table_in_array_variables_adding_axes = shape_array_variables_adding_axes[axis_variables_table_input]
    variables_table_in_array_variables_adding_axes = np.arange(n_variables_table_in_array_variables_adding_axes)

    if format_and_check:
        # check point 1
        if n_axes_array_variables_staying != n_axes_array_variables_adding_axes:
            raise ValueError('dimension mismatch')
        else:
            n_axes_array_input = n_axes_array_variables_staying

        if axis_variables_table_input < 0:
            axis_variables_table_input += n_axes_array_input
        if axis_samples_input < 0:
            axis_samples_input += n_axes_array_input

        # check point 6
        if axis_samples_input == axis_variables_table_input:
            raise ValueError('axis_samples_input and axis_variables_table_input must be different')

        # check point 2
        indexes_logical = np.arange(n_axes_array_input) != axis_variables_table_input
        if np.any(shape_array_variables_staying[indexes_logical] != shape_array_variables_adding_axes[indexes_logical]):
            raise ValueError('dimension mismatch')

        n_variables_table_in_array_variables_staying = shape_array_variables_staying[axis_variables_table_input]
        variables_table_in_array_variables_staying = np.arange(n_variables_table_in_array_variables_staying)

        n_variables_table_in_array_variables_adding_axes = shape_array_variables_adding_axes[axis_variables_table_input]
        variables_table_in_array_variables_adding_axes = np.arange(n_variables_table_in_array_variables_adding_axes)

        # format variables_table_adding_axes
        try:
            n_variables_table_adding_axes = len(variables_table_adding_axes)
            variables_table_adding_axes = np.asarray(variables_table_adding_axes, dtype=int)
            variables_table_adding_axes[variables_table_adding_axes < 0] += \
                n_variables_table_in_array_variables_adding_axes
            # check point 4
            if np.sum(variables_table_adding_axes[0] == variables_table_adding_axes) > 1:
                raise ValueError('variables_table_adding_axes cannot contain repeated values')
        except TypeError:
            if variables_table_adding_axes is None:
                variables_table_adding_axes = np.arange(n_variables_table_in_array_variables_adding_axes)
                n_variables_table_adding_axes = n_variables_table_in_array_variables_adding_axes
            else:
                if variables_table_adding_axes < 0:
                    variables_table_adding_axes += n_variables_table_in_array_variables_adding_axes
                variables_table_adding_axes = np.asarray([variables_table_adding_axes], dtype=int)
                n_variables_table_adding_axes = 1

        # format variables_table_staying
        try:
            n_variables_table_staying = len(variables_table_staying)
            variables_table_staying = np.asarray(variables_table_staying, dtype=int)
            variables_table_staying[variables_table_staying < 0] += n_variables_table_in_array_variables_staying
        except TypeError:
            if variables_table_staying is None:
                variables_table_staying = np.arange(n_variables_table_in_array_variables_staying)
                n_variables_table_staying = n_variables_table_in_array_variables_staying
            else:
                if variables_table_staying < 0:
                    variables_table_staying += n_variables_table_in_array_variables_staying
                variables_table_staying = np.asarray([variables_table_staying], dtype=int)
                n_variables_table_staying = 1

        # format axes_inserting
        n_axes_inserting = n_variables_table_adding_axes
        n_axes_array_output = n_axes_inserting
        try:
            n_axes_inserting = len(axes_inserting)
            axes_inserting = np.asarray(axes_inserting, dtype=int)
            axes_inserting[axes_inserting < 0] += n_axes_array_output
            # check point 3
            if np.sum(axes_inserting[0] == axes_inserting) > 1:
                raise ValueError('axes_inserting cannot contain repeated values')
            # check point 5
            if n_variables_table_adding_axes != n_axes_inserting:
                raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')
        except TypeError:
            if axes_inserting is None:
                axes_inserting = np.arange(n_axes_inserting)
            else:
                if axes_inserting < 0:
                    axes_inserting += n_axes_array_output
                axes_inserting = np.asarray([axes_inserting], dtype=int)
                n_axes_inserting = 1
                # check point 5
                if n_variables_table_adding_axes != n_axes_inserting:
                    raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')
    else:
        n_variables_table_staying = variables_table_staying.size
        n_variables_table_adding_axes = variables_table_adding_axes.size

    indexes = None
    if ((n_variables_table_staying != n_variables_table_in_array_variables_staying) or
            np.any(variables_table_staying != variables_table_in_array_variables_staying)):

        indexes = np.empty(n_axes_array_variables_staying, dtype=object)
        indexes[:] = slice(None)
        indexes[axis_variables_table_input] = variables_table_staying
        array_variables_staying = array_variables_staying[tuple(indexes)]

    if ((n_variables_table_adding_axes != n_variables_table_in_array_variables_adding_axes) or
            np.any(variables_table_adding_axes != variables_table_in_array_variables_adding_axes)):

        if indexes is None:
            indexes = np.empty(n_axes_array_variables_adding_axes, dtype=object)
            indexes[:] = slice(None)
        indexes[axis_variables_table_input] = variables_table_adding_axes
        array_variables_adding_axes = array_variables_adding_axes[tuple(indexes)]

    array_output = from_2_arrays(
        array_variables_staying,
        array_variables_adding_axes,
        axis_samples_input,
        axis_variables_table_input,
        axes_inserting=axes_inserting,
        dtype=dtype,
        format_and_check=False)

    return array_output


def from_1_array(
        array_input,
        axis_samples,
        axis_variables_table,
        variables_table_adding_axes,
        axes_inserting=None,
        variables_table_staying=None,
        dtype=None,
        format_and_check=True):

    # Notes:
    # 1) uses the efficient function slice for indexes;
    # 2) it splits the array in two arrays first: array_variables_adding_axes and array_variables_staying.
    #    Then, it makes the array output from those two arrays;
    # 3) the numbers of samples (or trials) in the axis_samples for all combinations
    #    of variables_table_adding_axes do not have to be equal;
    # 4) it does not assumes that the order of samples (or trials) in the axis_samples are is the same
    #    in each combination of variables' axes. The order is based on the order of the combinations
    #    of variables' conditions in axis_variables_table. In other words, the samples can be in random order
    #    (or in different order) in each combinations of variables axes/conditions.

    # Input requirements:
    # 1) axes_inserting cannot contain repeated values;
    # 2) variables_table_adding_axes cannot contain repeated values;
    # 3) shapes of axes_inserting and variables_table_adding_axes must be equal
    # 4) axis_samples != axis_variables_table

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.combinations import trials_to_conditions
    # 4) from ccalafiore.combinations import conditions_to_n_conditions
    # 5) from ccalafiore.combinations import conditions_to_combinations
    # 6) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # from ccalafiore.combinations import \
    #     trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations
    # from ccalafiore.array import samples_in_arr1_are_in_arr2

    axis_variables_table_input = axis_variables_table
    axis_samples_input = axis_samples

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = shape_array_input.size
    n_variables_table_input = shape_array_input[axis_variables_table_input]
    variables_table_input = np.arange(n_variables_table_input)

    if format_and_check:
        if axis_variables_table_input < 0:
            axis_variables_table_input += n_axes_array_input
        if axis_samples_input < 0:
            axis_samples_input += n_axes_array_input

        # check point 4
        if axis_samples_input == axis_variables_table_input:
            raise ValueError('axis_samples_input and axis_variables_table_input must be different')

        # format variables_table_adding_axes
        try:
            n_variables_table_adding_axes = len(variables_table_adding_axes)
            variables_table_adding_axes = np.asarray(variables_table_adding_axes, dtype=int)
            variables_table_adding_axes[variables_table_adding_axes < 0] += n_variables_table_input
            # check point 2
            if np.sum(variables_table_adding_axes[0] == variables_table_adding_axes) > 1:
                raise ValueError('variables_table_adding_axes cannot contain repeated values')
        except TypeError:
            if variables_table_adding_axes < 0:
                variables_table_adding_axes += n_variables_table_input
            variables_table_adding_axes = np.asarray([variables_table_adding_axes], dtype=int)
            n_variables_table_adding_axes = 1

        # format variables_table_staying
        try:
            n_variables_table_staying = len(variables_table_staying)
            variables_table_staying = np.asarray(variables_table_staying, dtype=int)
            variables_table_staying[variables_table_staying < 0] += n_variables_table_input
        except TypeError:
            if variables_table_staying is None:
                variables_table_staying = variables_table_input[np.logical_not(
                    samples_in_arr1_are_in_arr2(variables_table_input, variables_table_adding_axes))]
                n_variables_table_staying = variables_table_staying.size
            else:
                if variables_table_staying < 0:
                    variables_table_staying += n_variables_table_input
                variables_table_staying = np.asarray([variables_table_staying], dtype=int)
                n_variables_table_staying = 1

        # format axes_inserting
        n_axes_inserting = n_variables_table_adding_axes
        n_axes_array_output_object = n_axes_inserting
        try:
            n_axes_inserting = len(axes_inserting)
            axes_inserting = np.asarray(axes_inserting, dtype=int)
            axes_inserting[axes_inserting < 0] += n_axes_array_output_object
            # check point 1
            if np.sum(axes_inserting[0] == axes_inserting) > 1:
                raise ValueError('axes_inserting cannot contain repeated values')
            # check point 3
            if n_variables_table_adding_axes != n_axes_inserting:
                raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')
        except TypeError:
            if axes_inserting is None:
                axes_inserting = np.arange(n_axes_inserting)
            else:
                if axes_inserting < 0:
                    axes_inserting += n_axes_array_output_object
                axes_inserting = np.asarray([axes_inserting], dtype=int)
                n_axes_inserting = 1
                # check point 3
                if n_variables_table_adding_axes != n_axes_inserting:
                    raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')
    else:
        n_variables_table_staying = variables_table_staying.size
        n_variables_table_adding_axes = variables_table_adding_axes.size

    indexes = np.empty(n_axes_array_input, dtype=object)
    indexes[:] = slice(None)
    if ((n_variables_table_staying != n_variables_table_input) or
            np.any(variables_table_staying != variables_table_input)):
        indexes[axis_variables_table_input] = variables_table_staying
    array_variables_staying = array_input[tuple(indexes)]

    if ((n_variables_table_adding_axes != n_variables_table_input) or
            np.any(variables_table_adding_axes != variables_table_input)):
        indexes[axis_variables_table_input] = variables_table_adding_axes
    else:
        indexes[axis_variables_table_input] = slice(None)
    array_variables_adding_axes = array_input[tuple(indexes)]
    del array_input

    array_output = from_2_arrays(
        array_variables_staying,
        array_variables_adding_axes,
        axis_samples_input,
        axis_variables_table_input,
        axes_inserting=axes_inserting,
        dtype=dtype,
        format_and_check=False)

    return array_output



from matplotlib import pyplot as plt
# import seaborn as sns; sns.set()


def heat_maps(
        values, shape_figure=None,
        id_figures=None, directories_images=None, formats_images=['svg'], hspace=0.5,
        annot=False, fmt_annot=['.2g'], fontsize_annot=None,
        titles=None, x_labels=None, y_labels=None, xticklabels='auto', yticklabels='auto',
        fontsize_title='auto', rotation_title='auto',
        fontsize_x_labels='auto', rotation_x_labels='auto',
        fontsize_y_labels='auto', rotation_y_labels='auto',
        fontsize_xticklabels='auto', rotation_xticklabels='auto',
        fontsize_yticklabels='auto', rotation_yticklabels='auto',
        cmap=None, fontsize_cbar_ticklabels='auto'):

    figures_existing = plt.get_fignums()
    n_figures_existing = len(figures_existing)

    if id_figures is None:
        n_figures = values.shape[0]
        id_figures = [None] * n_figures
        i = 0
        f = 0
        while f < n_figures:
            if i in figures_existing:
                pass
            else:
                id_figures[f] = i
                f += 1
            i += 1
    else:
        n_figures = len(id_figures)
        for f in id_figures:
            if f in figures_existing:
                print('Warning: overwriting plot {}'.format(f))

    # plt.figure(0, frameon=False, dpi=200)
    # plt.figure(1,  frameon=False, dpi=200)
    if shape_figure is None:
        rows = values.shape[1]
        colomns = 1
    else:
        rows = shape_figure.shape[0]
        colomns = shape_figure.shape[1]

    n_figures = values.shape[0]
    n_subplots_per_fig = values.shape[1]

    a_num = ord('a')
    if titles is None:
        titles = ['{})'] * n_subplots_per_fig
    else:
        n_titles = len(titles)
        for t in range(n_titles):
            titles[t] = '{}) ' + titles[t]

    if cmap is None:
        cmap = [None] * n_subplots_per_fig

    if len(fmt_annot) == 1:
        fmt_annot = fmt_annot * n_subplots_per_fig
    # if x_labels is None:
    #     x_labels = [None] * n_subplots_per_fig
    #
    # if y_labels is None:
    #     y_labels = [None] * n_subplots_per_fig

    subplot = (100 * rows) + (10 * colomns) + 1
    for f in range(n_figures):

        plt.figure(id_figures[f], frameon=False, dpi=200)
        # plt.figure(f)

        for s in range(n_subplots_per_fig):

            plt.subplot(subplot + s)
            ax = sns.heatmap(
                values[f, s], annot=annot, fmt=fmt_annot[s], cmap=cmap[s],
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                annot_kws=dict([('fontsize', fontsize_annot)]))

            if xticklabels is not False:
                # r=0 and s=8
                if fontsize_xticklabels != 'auto' and rotation_xticklabels == 'auto':
                    plt.xticks(fontsize=fontsize_xticklabels)
                elif fontsize_xticklabels == 'auto' and rotation_xticklabels != 'auto':
                    plt.xticks(rotation=rotation_xticklabels)
                elif fontsize_xticklabels != 'auto' and rotation_xticklabels != 'auto':
                    plt.xticks(fontsize=fontsize_xticklabels, rotation=rotation_xticklabels)


            if yticklabels is not False:
                # r=0 and s=8
                if fontsize_yticklabels != 'auto' and rotation_yticklabels == 'auto':
                    plt.yticks(fontsize=fontsize_yticklabels)
                elif fontsize_yticklabels == 'auto' and rotation_yticklabels != 'auto':
                    plt.yticks(rotation=rotation_yticklabels)
                elif fontsize_yticklabels != 'auto' and rotation_yticklabels != 'auto':
                    plt.yticks(fontsize=fontsize_yticklabels, rotation=rotation_yticklabels)

            if fontsize_cbar_ticklabels != 'auto':
                # fontsize_cbar_ticklabels = 8
                # use matplotlib.colorbar.Colorbar object
                cbar = ax.collections[0].colorbar
                # here set the labelsize
                cbar.ax.tick_params(labelsize=fontsize_cbar_ticklabels)

            # general layout
            # fontsize_x_labels = 12
            # rotation_x_labels = 0
            if fontsize_title == 'auto' and rotation_title == 'auto':
                plt.title(titles[s].format(chr(a_num + s), f))
            elif fontsize_title != 'auto' and rotation_title == 'auto':
                plt.title(titles[s].format(chr(a_num + s), f), fontsize=fontsize_title)
            elif fontsize_title == 'auto' and rotation_title != 'auto':
                plt.title(titles[s].format(chr(a_num + s), f), rotation=rotation_title)
            elif fontsize_title != 'auto' and rotation_title != 'auto':
                plt.title(titles[s].format(chr(a_num + s), f), fontsize=fontsize_title, rotation=rotation_title)

            if x_labels is not None:
                # fontsize_x_labels = 10
                # rotation_x_labels = 0
                if fontsize_x_labels == 'auto' and rotation_x_labels == 'auto':
                    plt.xlabel(x_labels)
                elif fontsize_x_labels != 'auto' and rotation_x_labels == 'auto':
                    plt.xlabel(x_labels, fontsize=fontsize_x_labels)
                elif fontsize_x_labels == 'auto' and rotation_x_labels != 'auto':
                    plt.xlabel(x_labels, rotation=rotation_x_labels)
                elif fontsize_x_labels != 'auto' and rotation_x_labels != 'auto':
                    plt.xlabel(x_labels, fontsize=fontsize_x_labels, rotation=rotation_x_labels)

            if y_labels is not None:
                # fontsize_y_labels = 10
                # rotation_y_labels = 0
                if fontsize_y_labels == 'auto' and rotation_y_labels == 'auto':
                    plt.ylabel(y_labels)
                elif fontsize_y_labels != 'auto' and rotation_y_labels == 'auto':
                    plt.ylabel(y_labels, fontsize=fontsize_y_labels)
                elif fontsize_y_labels == 'auto' and rotation_y_labels != 'auto':
                    plt.ylabel(y_labels, rotation=rotation_y_labels)
                elif fontsize_y_labels != 'auto' and rotation_y_labels != 'auto':
                    plt.ylabel(y_labels, fontsize=fontsize_y_labels, rotation=rotation_y_labels)

    save_figure(id_figures=id_figures, directories=directories_images, formats=formats_images, hspace=hspace)

    plt.close('all')


def save_figure(id_figures=None, directories=None, formats=['svg'], hspace=0.5):
    
    if id_figures is None:
        id_figures = plt.get_fignums()
        
    n_figures = len(id_figures)

    if directories is None:
        directories = id_figures
    else:
        n_directories = len(directories)
        if n_figures != n_directories:
            raise ValueError('n_figures and n_directories must be equal.\n'
                             'Now, n_figures = {} and n_directories = {}'.format(n_figures, n_directories))

    n_formats = len(formats)
    for i in range(n_figures):

        plt.figure(id_figures[i])

        for j in range(n_formats):

            plt.subplots_adjust(hspace=hspace)
            directory_i_j = '.'.join([directories[i], formats[j]])
            plt.savefig(directory_i_j, format=formats[j], bbox_inches='tight', pad_inches=0.0)




__all__ = ['axes_to_variables_table', 'reorder_trials', 'reshape',
           'rotate_indexes', 'variables_table_to_axes']


def initiate(names_submodules=None):
    template_import_string = "from . import "
    if isinstance(names_submodules, str):
        import_string_m = template_import_string + names_submodules
        exec(import_string_m)
    elif isinstance(names_submodules, (list, tuple)):
        for name_m in names_submodules:
            if isinstance(name_m, str):
                import_string_m = template_import_string + name_m
            else:
                import_string_m = template_import_string + str(name_m)
            exec(import_string_m)
    elif names_submodules is None:
        exec('from . import *')
        # for name_m in __all__:
        #     if isinstance(name_m, str):
        #         import_string_m = template_import_string + name_m
        #     else:
        #         import_string_m = template_import_string + str(name_m)
        #     exec(import_string_m)
    else:
        raise TypeError(names_submodules)



# websites:
# https://pytorch.org/docs/stable/torchvision/transforms.html
# https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html#sphx-glr-beginner-blitz-cifar10-tutorial-py
# https://pytorch.org/hub/pytorch_vision_resnet/
# https://discuss.pytorch.org/t/normalize-each-input-image-in-a-batch-independently-and-inverse-normalize-the-output/23739
# https://pytorch.org/tutorials/beginner/transfer_learning_tutorial.html

import torch
import torchvision
import PIL
import math
import numpy as np
import os
import copy
from . import array as cc_array
from . import clock as cc_clock
from . import maths as cc_maths
from . import combinations as cc_combinations
from .strings import format_float_to_str as cc_strings_format_float_to_str


class Time:
    def __init__(self, range_raw, level, variate_shift=False, permute_shifts=False):
        self.range = cc_array.Range(range_raw)
        self.level = level

        self.permute_shifts = permute_shifts
        self.variate_shift = variate_shift

        if self.variate_shift:
            self.S = abs(self.range.step)
            if self.S == 1:
                self.variate_shift = False
                if self.permute_shifts:
                    self.permute_shifts = False
                self.shifts = np.asarray([0], dtype='i')

            elif self.permute_shifts:
                self.shifts = np.random.permutation(self.S)
            else:
                self.shifts = np.arange(self.S)
        else:
            self.S = 1
            self.shifts = np.asarray([0], dtype='i')

        self.s = 0
        self.shift = self.shifts[self.s]

    def next_shift(self):
        if self.variate_shift:
            if self.s < self.S - 1:
                self.s += 1
            else:
                self.s = 0
                if self.permute_shifts:
                    self.shifts = np.random.permutation(self.S)
            self.shift = self.shifts[self.s]


class Loader:

    def __init__(
            self, directory_root, level_classes, n_levels_directories=None,
            conditions_directories=None, batch_size=None, n_batches=None,
            shuffle=False, time=None, transforms=None, device=None):

        self.directory_root = directory_root
        self.level_classes = level_classes

        self.time = time

        if n_levels_directories is None:
            self.n_levels_directories = len(conditions_directories)
        elif conditions_directories is None:
            conditions_directories = [None] * self.n_levels_directories  # type: list

        self.L = self.n_levels_directories

        self.shuffle = shuffle

        if transforms is None:
            self.transforms = torchvision.transforms.Compose([torchvision.transforms.ToTensor()])
        else:
            self.transforms = transforms

        self.conditions_directories_names = [None] * self.L  # type: list
        self.n_conditions_directories = [None] * self.L  # type: list
        self.conditions_directories = [None] * self.L  # type: list

        directory_root_l = self.directory_root
        for l in range(self.L):
            self.conditions_directories_names[l] = os.listdir(directory_root_l)

            if conditions_directories[l] is None:
                self.n_conditions_directories[l] = len(self.conditions_directories_names[l])
                self.conditions_directories[l] = np.arange(self.n_conditions_directories[l])
            else:
                self.conditions_directories[l] = conditions_directories[l]
                self.n_conditions_directories[l] = len(self.conditions_directories[l])

            directory_root_l = os.path.join(directory_root_l, self.conditions_directories_names[l][0])

        if self.time is not None:
            # if self.time.T is None:
            #     self.time.T = self.n_conditions_directories[self.time.level]
            if self.time.range.stop is None:
                self.time.range.stop = self.n_conditions_directories[self.time.level]
            if self.time.level is None:
                self.time.range.level = self.L - 1

            self.conditions_directories[self.time.level] = (
                self.conditions_directories[self.time.level][self.time.range.to_slice()])
            self.n_conditions_directories[self.time.level] = len(self.conditions_directories[self.time.level])

        self.K = self.n_classes = self.n_conditions_directories[self.level_classes]
        self.n_samples = cc_maths.prod(self.n_conditions_directories)
        # self.n_samples = math.prod(self.n_conditions_directories)

        if batch_size is None:
            if n_batches is None:
                self.batch_size = self.n_samples
                self.n_batches = 1
            else:
                self.n_batches = n_batches
                self.batch_size = cc_maths.rint(self.n_samples / self.n_batches)
        else:
            self.batch_size = batch_size
            self.n_batches = cc_maths.rint(self.n_samples / self.batch_size)

        print('self.batch_size =', self.batch_size)
        print('self.n_samples =', self.n_samples)
        print('self.n_batches =', self.n_batches)
        print('self.time.range =', self.time.range)
        print('self.n_conditions_directories =', self.n_conditions_directories)
        # print('self.conditions_directories =', self.conditions_directories)

        if self.shuffle:
            self.batches_indexes = None
        else:
            self.batches_indexes = np.split(np.arange(self.n_samples), self.n_batches, axis=0)

        if self.time.variate_shift:
            self.combinations_directories_no_shift = (
                cc_combinations.conditions_to_combinations(self.conditions_directories))
            self.combinations_directories = np.copy(self.combinations_directories_no_shift)
            self.combinations_directories[:, self.time.level] += self.time.shift
        else:
            self.combinations_directories_no_shift = None
            self.combinations_directories = (
                cc_combinations.conditions_to_combinations(self.conditions_directories))

        self.labels = torch.tensor(
            self.combinations_directories[slice(0, self.n_samples, 1), self.level_classes],
            dtype=torch.int64, device=device)

        combination_directory_str_0 = [self.conditions_directories_names[v][0] for v in range(self.L)]
        directory_0 = os.path.join(self.directory_root, *combination_directory_str_0)
        image_0 = PIL.Image.open(directory_0)
        tensor_0 = self.transforms(image_0)
        # shape_sample_0 = np.asarray(tensor_0.shape)
        # self.shape_sample = shape_sample_0
        # self.shape_batch = np.empty(self.shape_sample.size + 1, self.shape_sample.dtype)
        # self.shape_batch[0] = self.batch_size
        # self.shape_batch[1:] = self.shape_sample
        # self.tensor_batch = torch.empty(tuple(self.shape_batch), dtype=tensor_0.dtype)
        shape_sample_0 = list(tensor_0.shape)
        self.shape_sample = shape_sample_0
        self.shape_batch = torch.Size(tuple([self.batch_size] + shape_sample_0))
        self.tensor_batch = torch.empty(self.shape_batch, dtype=tensor_0.dtype, device=device)

        self.n_dims_samples = self.shape_sample.__len__()
        self.n_dims_batch = self.n_dims_samples + 1
        self.indexes_batch = np.empty(self.n_dims_batch, dtype=object)
        self.indexes_batch[1:] = slice(0, None, 1)

        print('self.tensor_batch.shape =', self.tensor_batch.shape)
        print()

    def load_batches_e(self):

        if self.shuffle:
            self.batches_indexes = np.split(np.random.permutation(self.n_samples), self.n_batches, axis=0)

        # for b in range(3):
        for b in range(self.n_batches):
            labels_b = self.labels[self.batches_indexes[b]]
            # batch_size_eb = batches_indexes_e.shape[0]
            for i in range(self.batch_size):
                combination_directory_ebi = self.combinations_directories[self.batches_indexes[b][i], :]
                combination_directory_str_ebi = [
                    self.conditions_directories_names[l][combination_directory_ebi[l]]
                    for l in range(self.L)]
                directory_ebi = os.path.join(self.directory_root, *combination_directory_str_ebi)
                self.indexes_batch[0] = i

                image_ebi = PIL.Image.open(directory_ebi)
                self.tensor_batch[tuple(self.indexes_batch)] = self.transforms(image_ebi)

            yield [self.tensor_batch, labels_b]

        if (self.time is not None) and self.time.variate_shift:
            self.time.next_shift()
            self.combinations_directories = np.copy(self.combinations_directories_no_shift)
            self.combinations_directories[:, self.time.level] += self.time.shift


def train_early_stop(model, loader, criterion, optimizer, scheduler, I=10):

    # timer = cc_clock.Timer()
    cc_timer = cc_clock.Timer()

    for key_loader_k in loader.keys():
        if key_loader_k == 'training' or key_loader_k == 'validation':
            pass
        else:
            raise ValueError('Unknown keys in loader')

    headers = [
        'Epoch', 'Unsuccessful Epochs', 'Training Loss', 'Training Accuracy',
        'Valuation Loss', 'Lowest Valuation Loss', 'Is Lower Loss',
        'Valuation Accuracy', 'Highest Valuation Accuracy', 'Is Higher Accuracy']

    n_columns = len(headers)
    new_line_stats = [None] * n_columns  # type: list

    stats = {
        'headers': {headers[k]: k for k in range(n_columns)},
        'n_columns': n_columns,
        'lines': []}

    n_decimals_for_printing = 6

    best_model_wts = copy.deepcopy(model.state_dict())

    lowest_loss = math.inf
    lowest_loss_str = str(lowest_loss)
    highest_accuracy = -math.inf
    highest_accuracy_str = str(highest_accuracy)

    i = 0
    e = 0

    n_dashes = 110
    dashes = '-' * n_dashes
    print(dashes)

    while i < I:

        print('Epoch {} - Unsuccessful Epochs {}'.format(e, i))

        stats['lines'].append(new_line_stats.copy())
        stats['lines'][e][stats['headers']['Epoch']] = e
        stats['lines'][e][stats['headers']['Unsuccessful Epochs']] = i

        # Each epoch has a training and a validation phase
        # training phase
        model.train()  # Set model to training mode

        running_loss_e = 0.0
        running_corrects_e = 0

        b = 0
        # Iterate over data.
        for batch_eb, labels_eb in loader['training'].load_batches_e():

            # zero the parameter gradients
            optimizer.zero_grad()

            # forward
            # track history
            torch.set_grad_enabled(True)
            outputs = model(batch_eb)
            _, preds = torch.max(outputs, 1)
            loss_eb = criterion(outputs, labels_eb)

            # backward + optimize
            loss_eb.backward()
            optimizer.step()

            torch.set_grad_enabled(False)

            # # statistics
            # loss_float_eb = loss_eb.item()
            # # noinspection PyTypeChecker
            # corrects_eb = torch.sum(preds == labels_eb).item()
            # accuracy_eb = corrects_eb / batch_eb.shape[0]
            # loss_str_eb = cc_strings_format_float_to_str(loss_float_eb, n_decimals=n_decimals_for_printing)
            # accuracy_str_eb = cc_strings_format_float_to_str(accuracy_eb, n_decimals=n_decimals_for_printing)
            # print('Training. Epoch: {:d}. Batch {:d}. Loss: {:s}. Accuracy: {:s}.'.format(e, b, loss_str_eb, accuracy_str_eb))

            running_loss_e += loss_eb.item() * batch_eb.shape[0]
            # noinspection PyTypeChecker
            running_corrects_e += torch.sum(preds == labels_eb).item()

            b += 1

        # scheduler.step()

        loss_e = running_loss_e / loader['training'].n_samples
        accuracy_e = running_corrects_e / loader['training'].n_samples
        # loss_e = float(running_loss_e / loader['training'].n_samples)
        # accuracy_e = float(running_corrects_e / loader['training'].n_samples)

        loss_str_e = cc_strings_format_float_to_str(loss_e, n_decimals=n_decimals_for_printing)
        accuracy_str_e = cc_strings_format_float_to_str(accuracy_e, n_decimals=n_decimals_for_printing)

        print('Epoch: {:d}. Training.   Loss: {:s}. Accuracy: {:s}'.format(e, loss_str_e, accuracy_str_e))

        stats['lines'][e][stats['headers']['Training Loss']] = loss_e
        stats['lines'][e][stats['headers']['Training Accuracy']] = accuracy_e
        # stats['Training Loss'].append(float(loss_e))
        # stats['Training Accuracy'].append(float(accuracy_e))


        # validation phase
        model.eval()  # Set model to evaluate mode

        # zero the parameter gradients
        optimizer.zero_grad()

        torch.set_grad_enabled(False)

        running_loss_e = 0.0
        running_corrects_e = 0

        b = 0
        # Iterate over data.
        for batch_eb, labels_eb in loader['validation'].load_batches_e():

            # forward
            outputs = model(batch_eb)
            _, preds = torch.max(outputs, 1)
            loss_eb = criterion(outputs, labels_eb)

            # # statistics
            # loss_float_eb = loss_eb.item()
            # # noinspection PyTypeChecker
            # corrects_eb = torch.sum(preds == labels_eb).item()
            # accuracy_eb = corrects_eb / batch_eb.shape[0]
            # loss_str_eb = cc_strings_format_float_to_str(loss_float_eb, n_decimals=n_decimals_for_printing)
            # accuracy_str_eb = cc_strings_format_float_to_str(accuracy_eb, n_decimals=n_decimals_for_printing)
            # print('Validation. Epoch: {:d}. Batch {:d}. Loss: {:s}. Accuracy: {:s}.'.format(e, b, loss_str_eb, accuracy_str_eb))

            running_loss_e += loss_eb.item() * batch_eb.shape[0]
            # noinspection PyTypeChecker
            running_corrects_e += torch.sum(preds == labels_eb).item()

            b += 1

        loss_e = running_loss_e / loader['validation'].n_samples
        accuracy_e = running_corrects_e / loader['validation'].n_samples
        # loss_e = float(running_loss_e / loader['training'].n_samples)
        # accuracy_e = float(running_corrects_e / loader['training'].n_samples)

        loss_str_e = cc_strings_format_float_to_str(loss_e, n_decimals=n_decimals_for_printing)
        accuracy_str_e = cc_strings_format_float_to_str(accuracy_e, n_decimals=n_decimals_for_printing)

        print('Epoch: {:d}. Validation. Loss: {:s}. Lowest Loss {:s}. Accuracy: {:s}. Highest Accuracy: {:s}.'.format(
            e, loss_str_e, lowest_loss_str, accuracy_str_e, highest_accuracy_str))

        stats['lines'][e][stats['headers']['Valuation Loss']] = loss_e
        stats['lines'][e][stats['headers']['Lowest Valuation Loss']] = lowest_loss
        stats['lines'][e][stats['headers']['Valuation Accuracy']] = accuracy_e
        stats['lines'][e][stats['headers']['Highest Valuation Accuracy']] = highest_accuracy

        if loss_e < lowest_loss:
            stats['lines'][e][stats['headers']['Is Lower Loss']] = 1
            i = 0
            lowest_loss = loss_e
            lowest_loss_str = cc_strings_format_float_to_str(lowest_loss, n_decimals=n_decimals_for_printing)
            best_model_wts = copy.deepcopy(model.state_dict())  # deep copy the model
        else:
            stats['lines'][e][stats['headers']['Is Lower Loss']] = 0
            i += 1

        if accuracy_e > highest_accuracy:
            stats['lines'][e][stats['headers']['Is Higher Accuracy']] = 1
            highest_accuracy = accuracy_e
            highest_accuracy_str = cc_strings_format_float_to_str(highest_accuracy, n_decimals=n_decimals_for_printing)
        else:
            stats['lines'][e][stats['headers']['Is Higher Accuracy']] = 0

        e += 1
        print(dashes)

    print()
    E = e

    time_training = cc_timer.get_delta_time()

    print('Training complete in {d:s} days {h:s} hours {m:s} minutes {s:s} seconds'.format(
        d=str(time_training.days), h=str(time_training.hours),
        m=str(time_training.minutes), s=str(time_training.seconds)))
    print('Number of Epochs: {E:d}'.format(E=E))
    print('Lowest Validation Loss: {:s}'.format(lowest_loss_str))
    print('Highest Validation Accuracy: {:s}'.format(highest_accuracy_str))

    # load best model weights
    model.load_state_dict(best_model_wts)
    return model, stats


def load_model(name_model, K=None, pretrained=False, device=None):
    # model = torch.hub.load('pytorch/vision:v0.6.0', 'resnet152', pretrained=True)

    if isinstance(name_model, str):
        template_load_string = f"model = torchvision.models.{name_model:s}(pretrained=pretrained)"
    else:
        template_load_string = f"model = torchvision.models.{str(name_model):s}(pretrained=pretrained)"

    dict_globals = {'__builtins__': None}
    dict_locals = {'torchvision': torchvision, 'pretrained': pretrained}  # type: dict
    exec(template_load_string, dict_globals, dict_locals)
    model = dict_locals['model']  # type: torchvision.models.ResNet

    if K is not None:
        num_ftrs = model.fc.in_features
        # Here the size of each output sample is set to K.
        model.fc = torch.nn.Linear(num_ftrs, K)

    if device is not None:
        if isinstance(device, str):
            if device != 'cpu':
                model = model.to(torch.device(device))
        elif isinstance(device, torch.device):
            if device.type != 'cpu':
                model = model.to(device)

    return model



__version__ = '000.000.050'
__author__ = 'Calafiore Carmelo'
__author_email__ = 'c.calafiore@essex.ac.uk'
__maintainer_email__ = 'c.calafiore@essex.ac.uk'
__all__ = [
    'array', 'check_args', 'clock', 'combinations', 'directory', 'download',
    'format_args', 'image', 'maths', 'mixamo', 'nn', 'plot', 'preprocessing', 'shutdown', 'stats',
    'stimulation', 'strings', 'txt']


def initiate(names_submodules=None):
    template_import_string = "from . import "
    if isinstance(names_submodules, str):
        import_string_m = template_import_string + names_submodules
        exec(import_string_m)
    elif isinstance(names_submodules, (list, tuple)):
        for name_m in names_submodules:
            if isinstance(name_m, str):
                import_string_m = template_import_string + name_m
            else:
                import_string_m = template_import_string + str(name_m)
            exec(import_string_m)
    elif names_submodules is None:
        exec('from . import *')
        # for name_m in __all__:
        #     if isinstance(name_m, str):
        #         import_string_m = template_import_string + name_m
        #     else:
        #         import_string_m = template_import_string + str(name_m)
        #     exec(import_string_m)
    else:
        raise TypeError(names_submodules)



import os
import sys


def n_directories_up(directory, n=1):
    
    while n > 0:
        directory = os.path.dirname(directory)
        n -= 1

    return directory


def get_root_directory_of_running_script():
    directory_scripts = os.path.realpath(sys.argv[0])
    root_directory = os.path.dirname(directory_scripts)
    return root_directory



import numpy as np


def numeric_indexes_to_slice(indexes):

    n_axes = len(indexes)
    for a in range(n_axes):
        if not isinstance(indexes[a], slice):
            try:
                n_indexes_a = len(indexes[a])
                if n_indexes_a == 0:
                    pass
                elif n_indexes_a == 1:
                    # indexes[a] = slice(indexes[a][0], indexes[a][0] + 1, 1)
                    indexes[a] = indexes[a][0]
                elif n_indexes_a > 1:
                    sliceable = True
                    step = indexes[a][1] - indexes[a][0]
                    for i in range(2, n_indexes_a):
                        step_i = indexes[a][i] - indexes[a][i - 1]
                        if step_i != step:
                            sliceable = False
                            break
                    if sliceable:
                        if step > 0:
                            stop = indexes[a][-1] + 1
                        else:
                            stop = indexes[a][-1] - 1
                            if stop < 0:
                                stop = None
                        indexes[a] = slice(indexes[a][0], stop, step)

            except TypeError:
                pass
            # n_columns_views = 1

    return indexes


def format_shape_arguments(dict_arguments, target_shape):

    if isinstance(target_shape, list) or isinstance(target_shape, tuple):
        target_shape = np.asarray(target_shape, dtype=int)
    if isinstance(target_shape, np.ndarray):
        pass
    else:
        target_shape = np.asarray([target_shape], dtype=int)

    n_axes = target_shape.size

    # n_arguments = len(arguments)
    for key_i, argument_i in dict_arguments.items():

        # format arguments[i]
        if argument_i is None:
            argument_i = np.empty(target_shape, dtype=object)

        elif isinstance(argument_i, str):

            len_argument_i = len(argument_i)
            dtype_argument_i = '<U' + str(len_argument_i)
            argument_i_tmp = argument_i
            argument_i = np.empty(target_shape, dtype=dtype_argument_i)
            argument_i[:] = argument_i_tmp

        elif (isinstance(argument_i, list) or
              isinstance(argument_i, np.ndarray) or
              isinstance(argument_i, tuple)):

            if isinstance(argument_i, list) or isinstance(argument_i, tuple):
                # argument_i = np.asarray(argument_i)
                argument_i = np.asarray(argument_i)

            shape_argument_i = np.asarray(argument_i.shape, dtype=int)
            shape_in_argument_i = shape_argument_i[slice(0, n_axes, 1)]
            n_axis_in_argument_i = shape_in_argument_i.size
            if n_axis_in_argument_i != n_axes:
                shape_new_argument_i = np.append(target_shape, shape_argument_i)
                n_axis_in_new_argument_i = shape_new_argument_i.size
                dtype = argument_i.dtype
                argument_i_tmp = argument_i
                for a in range(n_axes):
                    argument_i_tmp = np.expand_dims(argument_i_tmp, axis=0)
                argument_i = np.empty(shape_new_argument_i, dtype=dtype)
                index = np.empty(n_axis_in_new_argument_i, dtype=object)
                index[:] = slice(None)
                argument_i[tuple(index)] = argument_i_tmp

            elif (shape_in_argument_i == target_shape).all():
                pass
            elif (shape_in_argument_i[shape_in_argument_i != target_shape] == 1).all():
                shape_argument_i = shape_argument_i[slice(n_axes, None, 1)]
                shape_new_argument_i = np.append(target_shape, shape_argument_i)
                n_axis_in_new_argument_i = shape_new_argument_i.size
                dtype = argument_i.dtype
                argument_i_tmp = argument_i
                argument_i = np.empty(shape_new_argument_i, dtype=dtype)
                index = np.empty(n_axis_in_new_argument_i, dtype=object)
                index[:] = slice(None)
                argument_i[tuple(index)] = argument_i_tmp[tuple(index)]
            else:
                shape_new_argument_i = np.append(target_shape, shape_argument_i)
                n_axis_in_new_argument_i = shape_new_argument_i.size
                dtype = argument_i.dtype
                argument_i_tmp = argument_i
                for a in range(n_axes):
                    argument_i_tmp = np.expand_dims(argument_i_tmp, axis=0)
                argument_i = np.empty(shape_new_argument_i, dtype=dtype)
                index = np.empty(n_axis_in_new_argument_i, dtype=object)
                index[:] = slice(None)
                argument_i[tuple(index)] = argument_i_tmp
        else:
            dtype_argument_i = type(argument_i)
            argument_i_tmp = argument_i
            argument_i = np.empty(target_shape, dtype=dtype_argument_i)
            argument_i[:] = argument_i_tmp

        dict_arguments[key_i] = argument_i

    return dict_arguments


def dict_to_key_array_and_value_array(dictionary):
    n_items = len(dictionary)
    keys = [None] * n_items
    values = np.empty(n_items, dtype=int)

    max_len_keys = 0
    i = 0
    for k, v in dictionary.items():
        keys[i] = k
        len_k = len(k)
        if len_k > max_len_keys:
            max_len_keys = len_k

        values[i] = v

        i += 1

    keys = np.asarray(keys, dtype=('<U' + str(max_len_keys)))
    return keys, values




import time as tm
import datetime
from . import maths


hours_per_day = 24
minutes_per_hour = 60
minutes_per_day = hours_per_day * minutes_per_hour
seconds_per_minute = 60
seconds_per_hour = minutes_per_hour * seconds_per_minute
seconds_per_day = hours_per_day * seconds_per_hour
milliseconds_per_second = 1000
milliseconds_per_minute = milliseconds_per_second * seconds_per_minute
milliseconds_per_hour = milliseconds_per_minute * minutes_per_hour
milliseconds_per_day = milliseconds_per_hour * hours_per_day
microseconds_per_millisecond = 1000
microseconds_per_second = microseconds_per_millisecond * milliseconds_per_second
microseconds_per_minute = microseconds_per_second * seconds_per_minute
microseconds_per_hour = microseconds_per_minute * minutes_per_hour
microseconds_per_day = microseconds_per_hour * hours_per_day


class SecTimer:

    def __init__(self):
        self.end_second = None
        self.delta_seconds = None
        self.start_second = tm.time()

    def define_delta_seconds(self):
        self.end_second = tm.time()
        self.delta_seconds = self.end_second - self.start_second

    def get_seconds(self):
        self.define_delta_seconds()
        return self.delta_seconds

    def get_milliseconds(self):
        self.define_delta_seconds()
        milliseconds = self.delta_seconds * milliseconds_per_second
        return milliseconds


class Timer:
    def __init__(self):
        self.end_datetime = None
        self.delta_time = None
        self.start_datetime = datetime.datetime.today()

    def define_delta_time(self):
        self.end_datetime = datetime.datetime.today()
        timedelta_builtin = self.end_datetime - self.start_datetime
        self.delta_time = TimeDelta(
            days=timedelta_builtin.days,
            seconds=timedelta_builtin.seconds,
            microseconds=timedelta_builtin.microseconds)

    def get_delta_time(self):
        self.define_delta_time()
        return self.delta_time

    def get_seconds(self):
        self.define_delta_time()
        seconds = self.delta_time.to_seconds()
        return seconds

    def get_milliseconds(self):
        self.define_delta_time()
        milliseconds = self.delta_time.to_milliseconds()
        return milliseconds


class TimeDelta:
    def __init__(self, days=0, hours=0, minutes=0, seconds=0, milliseconds=0, microseconds=0):
        self.type = 'day'

        self.microseconds = maths.convert_to_int_or_float(
            (maths.convert_to_int_or_float(days) * microseconds_per_day) +
            (maths.convert_to_int_or_float(hours) * microseconds_per_hour) +
            (maths.convert_to_int_or_float(minutes) * microseconds_per_minute) +
            (maths.convert_to_int_or_float(seconds) * microseconds_per_second) +
            (maths.convert_to_int_or_float(milliseconds) * microseconds_per_millisecond) +
            maths.convert_to_int_or_float(microseconds))

        self.milliseconds = int(self.microseconds / microseconds_per_millisecond)
        self.microseconds -= (self.milliseconds * microseconds_per_millisecond)

        self.seconds = int(self.milliseconds / milliseconds_per_second)
        self.milliseconds -= (self.seconds * milliseconds_per_second)

        self.minutes = int(self.seconds / seconds_per_minute)
        self.seconds -= (self.minutes * seconds_per_minute)

        self.hours = int(self.minutes / minutes_per_hour)
        self.minutes -= (self.hours * minutes_per_hour)

        self.days = int(self.hours / hours_per_day)
        self.hours -= (self.days * hours_per_day)

    def to_seconds(self):

        seconds = (
            (self.days * seconds_per_day) +
            (self.hours * seconds_per_hour) +
            (self.minutes * seconds_per_minute) +
            self.seconds +
            maths.convert_to_int_or_float(self.milliseconds / milliseconds_per_second) +
            maths.convert_to_int_or_float(self.microseconds / microseconds_per_second))

        return seconds

    def to_milliseconds(self):

        milliseconds = (
            (self.days * milliseconds_per_day) +
            (self.hours * milliseconds_per_hour) +
            (self.minutes * milliseconds_per_minute) +
            (self.seconds * milliseconds_per_second) +
            self.milliseconds +
            maths.convert_to_int_or_float(self.microseconds / microseconds_per_second))

        return milliseconds



import numpy as np
from ..combinations import n_conditions_to_combinations, conditions_to_combinations
from ..array import samples_in_arr1_are_in_arr2


def fast_balanced(
        array_input,
        axes_removing_input=0,
        axis_samples_input=-2,
        axis_variables_table_input=-1,
        variables_inserting_table_output=0,
        variables_staying_table_input=None,
        dtype=None):

    # Input requirements:
    # 1) axes_removing_input cannot contain same values;
    # 2) variables_inserting_table_output cannot contain same values;
    # 3) shapes of axes_removing_input and variables_inserting_table_output must be equal
    # 4) axes_removing_input[a] != axis_samples_input != axis_variables_table_input

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.array import samples_in_arr1_are_in_arr2
    # 4) from ccalafiore.array import advanced_indexing

    # from ccalafiore.array import samples_in_arr1_are_in_arr2, advanced_indexing

    # format axes_removing_input
    try:
        n_axes_removing_input = len(axes_removing_input)
        axes_removing_input = np.asarray(axes_removing_input, dtype=int)
    except TypeError:
        axes_removing_input = np.asarray([axes_removing_input], dtype=int)
        n_axes_removing_input = len(axes_removing_input)

    # format variables_inserting_in_axes_variables_output
    try:
        n_variables_inserting_table_output = len(variables_inserting_table_output)
        variables_inserting_table_output = np.asarray(variables_inserting_table_output, dtype=int)
    except TypeError:
        variables_inserting_table_output = np.asarray([variables_inserting_table_output], dtype=int)
        n_variables_inserting_table_output = len(variables_inserting_table_output)

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = len(shape_array_input)
    n_variables_table_input = shape_array_input[axis_variables_table_input]
    if axis_variables_table_input < 0:
        axis_variables_table_input += n_axes_array_input
    if axis_samples_input < 0:
        axis_samples_input += n_axes_array_input
    axes_removing_input[axes_removing_input < 0] += n_axes_array_input

    variables_table_input = np.arange(n_variables_table_input)
    if variables_staying_table_input is None:
        variables_staying_table_input = np.copy(variables_table_input)
    n_variables_staying_table_input = len(variables_staying_table_input)
    n_variables_table_output = n_variables_staying_table_input + n_variables_inserting_table_output
    variables_inserting_table_output[variables_inserting_table_output < 0] += n_variables_table_output

    # check point 1
    if np.sum(axes_removing_input[0] == axes_removing_input) > 1:
        raise Exception('axes_removing_input cannot contain repeated values')
    # check point 2
    if np.sum(
            variables_inserting_table_output[0] == variables_inserting_table_output) > 1:
        raise Exception('variables_inserting_table_output cannot contain repeated values')
    # check point 3
    if n_variables_inserting_table_output != n_axes_removing_input:
        raise Exception(
            'Shapes of axes_removing_input and variables_inserting_table_output must be equal')

    # check point 4
    if np.sum(axes_removing_input == axis_samples_input) > 0:
        raise Exception('axes_removing_input[a] != axis_samples_input')
    if np.sum(axes_removing_input == axis_variables_table_input) > 0:
        raise Exception('axes_removing_input[a] != axis_variables_table_input')
    if np.sum(axis_samples_input == axis_variables_table_input) > 0:
        raise Exception('axis_samples_input != axis_variables_table_input')

    axes_array_input = np.arange(n_axes_array_input)
    axes_other_array_input = axes_array_input[axes_array_input != axis_variables_table_input]
    axes_other_array_input = axes_other_array_input[
        axes_other_array_input != axis_samples_input]
    axes_other_array_input = axes_other_array_input[np.logical_not(
        samples_in_arr1_are_in_arr2(axes_other_array_input, axes_removing_input))]
    # axes_other_array_input_inverted = axes_other_array_input[::-1]
    # n_axes_other_array_input = len(axes_other_array_input)

    axis_variables_table_output = axis_variables_table_input - np.sum(axes_removing_input < axis_variables_table_input)
    axis_samples_output = axis_samples_input - np.sum(axes_removing_input < axis_samples_input)
    # changed = True
    # while changed:
    #     changed = False
    #     while axis_variables_table_output in axes_removing_input:
    #         axis_variables_table_output += 1
    #         changed = True
    #     while axis_samples_output in axes_removing_input:
    #         axis_samples_output += 1
    #         changed = True
    #     if axis_variables_table_output == axis_samples_output:
    #         changed = True
    #         if axis_variables_table_input > axis_samples_input:
    #             axis_variables_table_output += 1
    #         elif axis_variables_table_input < axis_samples_input:
    #             axis_samples_output += 1

    n_axes_array_output = n_axes_array_input - n_axes_removing_input
    axes_array_output = np.arange(n_axes_array_output)
    axes_non_axis_variables_table_output = axes_array_output[axes_array_output != axis_variables_table_output]
    axes_other_output = axes_non_axis_variables_table_output[
        axes_non_axis_variables_table_output != axis_samples_output]

    shape_array_output = np.empty(n_axes_array_output, dtype=object)
    n_conditions_in_axes_removing_input = shape_array_input[axes_removing_input]
    shape_array_output[axis_samples_output] = (
            shape_array_input[axis_samples_input] * np.prod(n_conditions_in_axes_removing_input))
    shape_array_output[axis_variables_table_output] = n_variables_table_output
    shape_array_output[axes_other_output] = shape_array_input[axes_other_array_input]

    if dtype is None:
        dtype = array_input.dtype
    array_output = np.empty(shape_array_output, dtype=dtype)

    variables_table_output = np.arange(n_variables_table_output)
    variables_staying_table_output = variables_table_output[np.logical_not(
        samples_in_arr1_are_in_arr2(variables_table_output, variables_inserting_table_output))]

    axis_variables_in_combinations_removing = int(axis_variables_table_input > axis_samples_input)
    axis_combinations_in_combinations_removing = int(not (bool(axis_variables_in_combinations_removing)))
    combinations_axes_removing_input = n_conditions_to_combinations(
        n_conditions_in_axes_removing_input, axis_combinations=axis_combinations_in_combinations_removing)
    n_combinations_axes_removing_input = (
        combinations_axes_removing_input.shape[axis_combinations_in_combinations_removing])

    indexes_combinations_removing_input = np.empty(2, dtype=object)
    indexes_combinations_removing_input[axis_variables_in_combinations_removing] = slice(None)

    indexes_combinations_removing_output = np.empty(2, dtype=object)
    indexes_combinations_removing_output[axis_variables_in_combinations_removing] = np.arange(n_axes_removing_input)

    # for a in axes_other_output:
    #     combinations_axes_removing_input = np.expand_dims(combinations_axes_removing_input, axis=a)



    axes_removing_input_inverted = np.sort(axes_removing_input)[::-1]

    # index_input = np.full(n_axes_array_input, slice(None), dtype=object)
    indexes_input = np.empty(n_axes_array_input, dtype=object)
    for a in range(n_axes_array_input):
        if (a not in axes_removing_input) and (a != axis_variables_table_input):
            indexes_input[a] = np.arange(shape_array_input[a])

    indexes_input[axis_variables_table_input] = variables_staying_table_input

    # index_array_output = np.full(n_axes_array_output, slice(None), dtype=object)
    index_array_output = np.empty(n_axes_array_output, dtype=object)
    for a in range(n_axes_array_output):
        if (a != axis_variables_table_output) and (a != axis_samples_output):
            index_array_output[a] = np.arange(shape_array_output[a])

    length_axis_flattering_input = shape_array_input[axis_samples_input]

    start_index_axis_samples_output = 0
    for c in range(n_combinations_axes_removing_input):

        stop_index_axis_samples_output = (c + 1) * length_axis_flattering_input
        index_array_output[axis_samples_output] = np.arange(
            start_index_axis_samples_output, stop_index_axis_samples_output)
        start_index_axis_samples_output = stop_index_axis_samples_output

        index_array_output[axis_variables_table_output] = variables_staying_table_output

        indexes_combinations_removing_input[axis_combinations_in_combinations_removing] = c
        indexes_input[axes_removing_input] = (
            combinations_axes_removing_input[tuple(indexes_combinations_removing_input)])
        array_input_c = array_input[advanced_indexing(indexes_input)]
        for a in axes_removing_input_inverted:
            array_input_c = np.squeeze(array_input_c, axis=a)

        array_output[advanced_indexing(index_array_output)] = array_input_c
        index_array_output[axis_variables_table_output] = variables_inserting_table_output

        indexes_combinations_removing_output[axis_combinations_in_combinations_removing] = c
        combinations_axes_removing_input_c = combinations_axes_removing_input[
            advanced_indexing(indexes_combinations_removing_output)]
        for a in axes_other_output:
            combinations_axes_removing_input_c = np.expand_dims(combinations_axes_removing_input_c, axis=a)

        array_output[advanced_indexing(index_array_output)] = combinations_axes_removing_input_c

    return array_output
